"""FIFE: Finite-Interval Forecasting Engine.

Copyright (c) 2018 - 2021, Institute for Defense Analyses (IDA)

Please cite using the suggested citation in the LICENSE file.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU Affero General Public License for more details
(a copy is included in the LICENSE file).
"""

__version__ = "1.5.2"



"""Data processing functions and classes for FIFE."""

from typing import List, Tuple, Union

import dask
import numpy as np
import pandas as pd

from warnings import warn


def check_column_consistency(data: pd.core.frame.DataFrame, colname: str) -> None:
    """Assert column exists, has no missing values, and is not constant."""
    assert colname in data, f"{colname} not in data"
    assert all(data[colname].notnull()), f"{colname} has missing values"
    assert any(
        data[colname][1:].values != data[colname][:-1].values
    ), f"{colname} does not have multiple unique values"


def deduplicate_column_values(
    data: pd.core.frame.DataFrame, reserved_cols: List[str] = [], max_obs: int = 65536
) -> pd.core.frame.DataFrame:
    """Delete columns with the same values as a later column.

    Args:
        df: A DataFrame.
        reserved_cols: Names of columns to exclude from deduplication.
        max_obs: The number of observations to sample if df has more than that
            many observations.

    Returns:
        A DataFrame containing only the last instance of each unique column.
    """
    warn(
        "Warning: deduplicate_column_values is deprecated. "
        "If you wish to deduplicate columns, do so prior to initializing a processor.",
        DeprecationWarning,
    )
    comparison_data = data.drop(reserved_cols, axis=1).sample(
        n=min(max_obs, data.shape[0]), replace=False
    )
    deduplicated_cols = list(comparison_data.T.drop_duplicates(keep="last").index)
    deduplicated_data = data[reserved_cols + deduplicated_cols]
    duplicated_cols = [col for col in data if col not in deduplicated_data.columns]
    if duplicated_cols:
        print(
            f'{", ".join(duplicated_cols)} dropped for having identical '
            + "values as another feature"
        )
    return deduplicated_data


def normalize_numeric_feature(
    col: pd.core.series.Series, excluded_obs: Union[None, pd.core.series.Series] = None
) -> Tuple[pd.core.series.Series, List[float]]:
    """Scale numeric values to their empirical range.

    Args:
        col: A numeric Series.
        excluded_obs: True for observations to exclude when computing min/max.

    Returns:
        - A Series of floats.
        - A list containing the minimum and maximum values among the included
          observations.
    """
    warn(
        "Warning: normalize_numeric_feature is deprecated. "
        "Modelers now use transform_features upon initialization instead.",
        DeprecationWarning,
    )
    if excluded_obs is None:
        col_subset = col[col.notnull()]
    else:
        col_subset = col[(~excluded_obs) & col.notnull()]
    if len(col_subset) > 0:
        minimum = np.nanmin(col_subset)
        maximum = np.nanmax(col_subset)
    else:
        minimum = np.nan
        maximum = np.nan
    normalized_col = process_numeric_feature(col, minimum, maximum)
    numeric_range = [minimum, maximum]
    return normalized_col, numeric_range


def process_numeric_feature(
    col: pd.core.series.Series, minimum: float, maximum: float
) -> pd.core.series.Series:
    """Scale numeric values to a given range.

    Args:
        col: A numeric Series.
        minimum: The value to map to -0.5.
        maximum: The value to map to 0.5.

    Returns:
        A Series of floats.
    """
    warn(
        "Warning: process_numeric_feature is deprecated. "
        "Modelers now use transform_features upon initialization instead.",
        DeprecationWarning,
    )
    return (col - minimum) / (maximum - minimum) - 0.5


def factorize_categorical_feature(
    col: pd.core.series.Series, excluded_obs: Union[None, pd.core.series.Series] = None
) -> Tuple[pd.core.series.Series, dict]:
    """Map categorical values to unsigned integers.

    Args:
        col: A Series.
        excluded_obs: True for observations to exclude from creating the map.

    Returns:
        - A pandas Series of unsigned integers.
        - A dictionary mapping each unique value among the included
          observations and np.nan to an integer and any other value not in the
          Series to 0.
    """
    warn(
        "Warning: factorize_categorical_feature is deprecated. "
        "Processors now convert categorical features to pandas Categorical type.",
        DeprecationWarning,
    )
    if excluded_obs is None:
        cat_map = produce_categorical_map(col)
    else:
        cat_map = produce_categorical_map(col[~excluded_obs])
    factorized_col = process_categorical_feature(col, cat_map)
    return factorized_col, cat_map


def produce_categorical_map(col: pd.core.series.Series) -> dict:
    """Return a map from categorical values to unsigned integers.

    Zero is reserved for values not seen in data used to create map.

    Args:
        col: A Series.

    Returns:
        A dictionary mapping each unique value in the Series and np.nan to an
        integer and any other value not in the Series to zero.
    """
    warn(
        "Warning: produce_categorical_map is deprecated. "
        "Processors now convert categorical features to pandas Categorical type.",
        DeprecationWarning,
    )
    _, cat_map = pd.factorize(col.values, sort=True)
    cat_map = {val: (i + 1) for i, val in enumerate(cat_map)}
    cat_map["unrecognized"] = 0
    cat_map[np.nan] = len(cat_map)
    return cat_map


def process_categorical_feature(
    col: pd.core.series.Series, cat_map: dict
) -> pd.core.frame.DataFrame:
    """Map categorical values to unsigned integers.

    Args:
        col: A pandas Series.
        cat_map: A dict containing a key for each unique value in col.

    Returns:
        A pandas Series of unsigned integers.
    """
    warn(
        "Warning: process_categorical_feature is deprecated. "
        "Processors now convert categorical features to pandas Categorical type.",
        DeprecationWarning,
    )
    col = col.map(cat_map).fillna(0)
    n_bits = 8
    while col.max() >= 2 ** n_bits:
        n_bits *= 2
    col = col.astype("uint" + str(n_bits))
    return col


class DataProcessor:
    """Prepare data by identifying features as degenerate or categorical."""

    def __init__(
        self,
        config: Union[None, dict] = {},
        data: Union[None, pd.core.frame.DataFrame] = None,
    ) -> None:
        """Initialize the DataProcessor.

        Args:
            config: A dictionary of configuration parameters.
            data: A DataFrame to be processed.
        """
        if (config.get("INDIVIDUAL_IDENTIFIER", "") == "") and data is not None:
            config["INDIVIDUAL_IDENTIFIER"] = data.columns[0]
            print(
                "Individual identifier column name not given; assumed to be "
                f'leftmost column ({config["INDIVIDUAL_IDENTIFIER"]})'
            )
        self.config = config
        self.data = data

    def is_degenerate(self, col: str) -> bool:
        """Determine if a feature is constant or has too many missing values."""
        if self.data[col].isnull().mean() >= self.config.get("MAX_NULL_SHARE", 0.999):
            return True
        if self.data[col].nunique(dropna=False) < 2:
            return True
        return False

    def is_categorical(self, col: str) -> bool:
        """Determine if the given feature should be processed as categorical, as opposed to numeric."""
        if col.endswith(tuple(self.config.get("CATEGORICAL_SUFFIXES", []))):
            if col.endswith(tuple(self.config.get("NUMERIC_SUFFIXES", []))):
                print(
                    f"{col} matches categorical and numeric suffixes; "
                    "identified as categorical"
                )
            return True
        if col in self.data.select_dtypes(
            include=["datetime"]
        ) or col == self.config.get("TIME_IDENTIFIER"):
            return False
        if col in self.data.select_dtypes(exclude=["number"]):
            if col.endswith(tuple(self.config.get("NUMERIC_SUFFIXES", []))):
                print(
                    f"{col} matches numeric suffix but is non-numeric; "
                    "identified as categorical"
                )
            return True
        if col.endswith(tuple(self.config.get("NUMERIC_SUFFIXES", []))) or (
            self.data[col].nunique() > self.config.get("MAX_UNIQUE_NUMERIC_CATS", 1024)
        ):
            return False
        return True


class PanelDataProcessor(DataProcessor):
    """Ready panel data for modelling.

    Attributes:
        config (dict): User-provided configuration parameters.
        data (pd.core.frame.DataFrame): Processed panel data.
        raw_subset (pd.core.frame.DataFrame): An unprocessed sample from the
            final period of data. Useful for displaying meaningful values in
            SHAP plots.
        categorical_maps (dict): Contains for each categorical feature a map
            from each unique value to a whole number.
        numeric_ranges (pd.core.frame.DataFrame): Contains for each numeric
            feature the maximum and minimum value in the training set.
    """

    def __init__(
        self,
        config: Union[None, dict] = {},
        data: Union[None, pd.core.frame.DataFrame] = None,
    ) -> None:
        """Initialize the PanelDataProcessor.

        Args:
            config: A dictionary of configuration parameters.
            data: A DataFrame to be processed.
        """
        if (config.get("TIME_IDENTIFIER", "") == "") and data is not None:
            config["TIME_IDENTIFIER"] = data.columns[1]
            print(
                "Time identifier column name not given; assumed to be "
                f'second-leftmost column ({config["TIME_IDENTIFIER"]})'
            )
        super().__init__(config, data)

    def build_processed_data(self, parallelize: bool = True) -> None:
        """Clean, augment, and store a panel dataset and related information.

        - Sort data by individual and time.
        - Drop degenerate features.
        - Label subsets for prediction, validation, and testing.
        - Compute survival duration and if departure is observed.
        - Store a subset of the raw input data from the final period.
        - Map categorical features to unsigned integers.
        - Scale numeric features.
        """
        self.check_panel_consistency()
        self.data = self.sort_panel_data()
        self.process_all_columns(parallelize=parallelize)
        self.build_reserved_cols()

    def process_all_columns(self, parallelize: bool = True) -> None:
        """Split, process, and merge all data columns."""
        data_dict = {}
        if parallelize:
            for colname in self.data:
                data_dict[colname] = dask.delayed(self.process_single_column)(colname)
            data_dict = dask.compute(data_dict)[0]
        else:
            for colname in self.data:
                data_dict[colname] = self.process_single_column(colname)
        data_dict = {key: val for key, val in data_dict.items() if val is not None}
        self.data = pd.DataFrame.from_dict(data_dict)

    def process_single_column(self, colname: str) -> Union[None, pd.core.series.Series]:
        """Apply data cleaning functions to an individual data column."""
        if colname == self.config["INDIVIDUAL_IDENTIFIER"]:
            return self.data[colname]
        elif self.is_degenerate(colname):
            print(f'Column "{colname}" is degenerate and will be dropped.')
            return None
        elif self.is_categorical(colname):
            return (
                self.data[colname]
                .astype("category")
                .cat.add_categories("NaN")
                .fillna("NaN")
            )
        else:
            return self.data[colname]

    def build_reserved_cols(self):
        """Add data split and outcome-related columns to the data."""
        self.data["_period"] = pd.factorize(
            self.data[self.config["TIME_IDENTIFIER"]], sort=True
        )[0]
        self.data["_predict_obs"] = self.data["_period"] == self.data["_period"].max()

        max_test_intervals = int((len(set(self.data["_period"])) - 1) / 2)
        test_intervals = self.config.get(
            "TEST_INTERVALS", self.config.get("TEST_PERIODS", 0) - 1
        )

        if test_intervals > max_test_intervals:
            warn(
                f"The specified value for TEST_INTERVALS of {test_intervals} was too high and will not allow for enough training periods. It was automatically reduced to {max_test_intervals}"
            )
            test_intervals = max_test_intervals

        self.data["_test"] = (self.data["_period"] + test_intervals) >= self.data[
            "_period"
        ].max()
        self.data["_validation"] = (
            self.flag_validation_individuals() & ~self.data["_test"]
        )
        observation_max_period = self.data.groupby("_test")["_period"].transform("max")
        self.data["_maximum_lead"] = (
            observation_max_period
            - self.data["_period"]
            + (observation_max_period < self.data["_period"].max())
        )
        del observation_max_period
        gaps = (
            self.data.groupby(self.config["INDIVIDUAL_IDENTIFIER"])["_period"].shift()
            < self.data["_period"] - 1
        )
        spells = gaps.groupby(self.data[self.config["INDIVIDUAL_IDENTIFIER"]]).cumsum()
        self.data["_duration"] = spells.groupby(
            [self.data[self.config["INDIVIDUAL_IDENTIFIER"]], spells]
        ).cumcount(ascending=False)
        self.data["_spell"] = spells
        self.data["_event_observed"] = (
            self.data["_duration"] < self.data["_maximum_lead"]
        )

    def check_panel_consistency(self) -> None:
        """Ensure observations have unique individual-period combinations."""
        check_column_consistency(self.data, self.config["INDIVIDUAL_IDENTIFIER"])
        check_column_consistency(self.data, self.config["TIME_IDENTIFIER"])
        ids = [self.config["INDIVIDUAL_IDENTIFIER"], self.config["TIME_IDENTIFIER"]]
        assert not any(self.data.duplicated(subset=ids)), (
            "One or more individuals have multiple observations for "
            "a single time value"
        )

    def sort_panel_data(self) -> pd.core.frame.DataFrame:
        """Sort the data by individual, then by period."""
        return self.data.sort_values(
            [self.config["INDIVIDUAL_IDENTIFIER"], self.config["TIME_IDENTIFIER"]]
        ).reset_index(drop=True)

    def flag_validation_individuals(self) -> pd.core.series.Series:
        """Flag observations from a random share of individuals."""
        unique_ids = self.data[self.config["INDIVIDUAL_IDENTIFIER"]].unique()
        size = int(self.config.get("VALIDATION_SHARE", 0.25) * unique_ids.shape[0])
        validation_ids = np.random.choice(unique_ids, size=size, replace=False)
        return self.data[self.config["INDIVIDUAL_IDENTIFIER"]].isin(validation_ids)



# nnet_survival.py
#
# Functions used in neural network survival model.
# To use in your program, either go the directory containing nnet_survival.py and then start Python,
# or add that directory to Python path.
#
# Author: Michael Gensheimer, Stanford University, michael.gensheimer@gmail.com
# Tested with Python version 3.6 and Keras version 2 (using TensorFlow backend)

from __future__ import print_function
import numpy as np
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Layer


def surv_likelihood(n_intervals):
    """Create custom Keras loss function for neural network survival model.

    Arguments
        n_intervals: the number of survival time intervals
    Returns
        Custom loss function that can be used with Keras
    """

    def loss(y_true, y_pred):
        """
        Required to have only 2 arguments by Keras.

        Arguments
            y_true: Tensor.
              First half of the values is 1 if individual survived that interval, 0 if not.
              Second half of the values is for individuals who failed, and is 1 for time interval during which failure occured, 0 for other intervals.
              See make_surv_array function.
            y_pred: Tensor, predicted survival probability (1-hazard probability) for each time interval.
        Returns
            Vector of losses for this minibatch.
        """

        cens_uncens = 1.0 + y_true[:, 0:n_intervals] * (
            y_pred - 1.0
        )  # component for all individuals
        uncens = (
            1.0 - y_true[:, n_intervals : 2 * n_intervals] * y_pred
        )  # component for only uncensored individuals
        return K.sum(
            -K.log(K.clip(K.concatenate((cens_uncens, uncens)), K.epsilon(), None)),
            axis=-1,
        )  # return -log likelihood

    return loss


def surv_likelihood_rnn(n_intervals):
    """Create custom Keras loss function for neural network survival model. Used for recurrent neural networks with time-distributed output.
    This function is very similar to surv_likelihood but deals with the extra dimension of y_true and y_pred that exists because of the time-distributed output.
    """

    def loss(y_true, y_pred):
        cens_uncens = 1.0 + y_true[0, :, 0:n_intervals] * (
            y_pred - 1.0
        )  # component for all patients
        uncens = (
            1.0 - y_true[0, :, n_intervals : 2 * n_intervals] * y_pred
        )  # component for only uncensored patients
        return K.sum(
            -K.log(K.clip(K.concatenate((cens_uncens, uncens)), K.epsilon(), None)),
            axis=-1,
        )  # return -log likelihood

    return loss


def make_surv_array(t, f, breaks):
    """Transforms censored survival data into vector format that can be used in Keras.

    Arguments
        t: Array of failure/censoring times.
        f: Censoring indicator. 1 if failed, 0 if censored.
        breaks: Locations of breaks between time intervals for discrete-time survival model (always includes 0)
    Returns
        Two-dimensional array of survival data, dimensions are number of individuals X number of time intervals*2
    """

    n_samples = t.shape[0]
    n_intervals = len(breaks) - 1
    timegap = breaks[1:] - breaks[:-1]
    breaks_midpoint = breaks[:-1] + 0.5 * timegap
    y_train = np.zeros((n_samples, n_intervals * 2))
    for i in range(n_samples):
        if f[i]:  # if failed (not censored)
            y_train[i, 0:n_intervals] = 1.0 * (
                t[i] >= breaks[1:]
            )  # give credit for surviving each time interval where failure time >= upper limit
            if (
                t[i] < breaks[-1]
            ):  # if failure time is greater than end of last time interval, no time interval will have failure marked
                y_train[
                    i, n_intervals + np.where(t[i] < breaks[1:])[0][0]
                ] = 1  # mark failure at first bin where survival time < upper break-point
        else:  # if censored
            y_train[i, 0:n_intervals] = 1.0 * (
                t[i] >= breaks_midpoint
            )  # if censored and lived more than half-way through interval, give credit for surviving the interval.
    return y_train


def nnet_pred_surv(y_pred, breaks, fu_time):
    # Predicted survival probability from Nnet-survival model
    # Inputs are Numpy arrays.
    # y_pred: Rectangular array, each individual's conditional probability of surviving each time interval
    # breaks: Break-points for time intervals used for Nnet-survival model, starting with 0
    # fu_time: Follow-up time point at which predictions are needed
    #
    # Returns: predicted survival probability for each individual at specified follow-up time
    y_pred = np.cumprod(y_pred, axis=1)
    pred_surv = []
    for i in range(y_pred.shape[0]):
        pred_surv.append(np.interp(fu_time, breaks[1:], y_pred[i, :]))
    return np.array(pred_surv)


class PropHazards(Layer):
    def __init__(self, output_dim, **kwargs):
        self.output_dim = output_dim
        super(PropHazards, self).__init__(**kwargs)

    def build(self, input_shape):
        # Create a trainable weight variable for this layer.
        self.kernel = self.add_weight(
            name="kernel",
            shape=(1, self.output_dim),
            # initializer='uniform',
            initializer="zeros",
            trainable=True,
        )
        super(PropHazards, self).build(input_shape)  # Be sure to call this somewhere!

    def call(self, x):
        # The conditional probability of surviving each time interval (given that has survived to beginning of interval)
        # is affected by the input data according to eq. 18.13 in Harrell F.,
        # Regression Modeling Strategies 2nd ed. (available free online)
        return K.pow(K.sigmoid(self.kernel), K.exp(x))

    def compute_output_shape(self, input_shape):
        return (input_shape[0], self.output_dim)

    def get_config(self):
        config = super().get_config().copy()
        config["output_dim"] = self.output_dim
        return config
